@@ -20,41 +20,51 @@ import tempfile
 import threading
 import webbrowser
 
+from .tornado.kernel_websocket_handler import VoilaKernelWebsocketHandler
+
+from .tornado.execution_request_handler import ExecutionRequestHandler
+
 from .tornado.contentshandler import VoilaContentsHandler
 
-from .voila_identity_provider import VoilaLoginHandler
-
-try:
-    from urllib.parse import urljoin
-    from urllib.request import pathname2url
-except ImportError:
-    from urllib import pathname2url
-
-    from urlparse import urljoin
+from urllib.parse import urljoin
+from urllib.request import pathname2url
 
 import jinja2
 import tornado.ioloop
 import tornado.web
-from jupyter_core.paths import jupyter_config_path, jupyter_path
+from jupyter_core.paths import jupyter_config_path
 from jupyter_server.base.handlers import FileFindHandler, path_regex
 from jupyter_server.config_manager import recursive_update
-from jupyter_server.services.config.manager import ConfigManager
 from jupyter_server.services.contents.largefilemanager import LargeFileManager
 from jupyter_server.services.kernels.handlers import KernelHandler
-from jupyter_server.services.kernels.websocket import KernelWebsocketHandler
-from jupyter_server.auth.authorizer import AllowAllAuthorizer, Authorizer
-from jupyter_server.auth.identity import PasswordIdentityProvider
-from jupyter_server import DEFAULT_TEMPLATE_PATH_LIST, DEFAULT_STATIC_FILES_PATH
-from jupyter_server.services.kernels.connection.base import (
-    BaseKernelWebsocketConnection,
-)
-from jupyter_server.services.kernels.connection.channels import (
-    ZMQChannelsWebsocketConnection,
-)
-from jupyter_server.auth.identity import (
-    IdentityProvider,
-)
-from jupyter_server.utils import url_path_join
+
+try:
+    JUPYTER_SERVER_2 = True
+
+    from jupyter_server.auth.authorizer import AllowAllAuthorizer, Authorizer
+    from jupyter_server.auth.identity import PasswordIdentityProvider
+    from jupyter_server import DEFAULT_TEMPLATE_PATH_LIST, DEFAULT_STATIC_FILES_PATH
+    from jupyter_server.services.kernels.connection.base import (
+        BaseKernelWebsocketConnection,
+    )
+    from jupyter_server.services.kernels.connection.channels import (
+        ZMQChannelsWebsocketConnection,
+    )
+    from jupyter_server.auth.identity import (
+        IdentityProvider,
+    )
+    from jupyter_server.utils import url_path_join
+    from jupyter_core.utils import run_sync
+    from jupyter_server.services.config.manager import ConfigManager
+
+    from jupyterlab_server.themes_handler import ThemesHandler
+    from .voila_identity_provider import VoilaLoginHandler
+except ImportError:
+    JUPYTER_SERVER_2 = False
+
+    from jupyter_server.utils import url_path_join, run_sync
+    from jupyter_server.services.config import ConfigManager
+
 from jupyter_core.utils import run_sync
 
 from jupyterlab_server.themes_handler import ThemesHandler
@@ -90,7 +100,12 @@ from .static_file_handler import (
 )
 from .tornado.handler import TornadoVoilaHandler
 from .tornado.treehandler import TornadoVoilaTreeHandler
-from .utils import create_include_assets_functions, get_data_dir, pjoin
+from .utils import (
+    create_include_assets_functions,
+    get_data_dir,
+    pjoin,
+    get_voila_labextensions_path,
+)
 from .voila_kernel_manager import voila_kernel_manager_factory
 
 _kernel_id_regex = r"(?P<kernel_id>\w+-\w+-\w+-\w+-\w+)"
@@ -127,10 +142,6 @@ class Voila(Application):
                 'Show left and right margins for the "lab" template, this gives a "classic" template look'
             ),
         ),
-        "token": (
-            {"Voila": {"auto_token": True}},
-            _(""),
-        ),
         "classic-tree": (
             {
                 "VoilaConfiguration": {"classic_tree": True},
@@ -139,6 +150,15 @@ class Voila(Application):
         ),
     }
 
+    if JUPYTER_SERVER_2:
+        flags = {
+            **flags,
+            "token": (
+                {"Voila": {"auto_token": True}},
+                _(""),
+            ),
+        }
+
     description = Unicode(
         """voila [OPTIONS] NOTEBOOK_FILENAME
 
@@ -172,16 +192,18 @@ class Voila(Application):
         "port": "Voila.port",
         "static": "Voila.static_root",
         "server_url": "Voila.server_url",
-        "token": "Voila.token",
         "pool_size": "VoilaConfiguration.default_pool_size",
         "show_tracebacks": "VoilaConfiguration.show_tracebacks",
         "preheat_kernel": "VoilaConfiguration.preheat_kernel",
+        "progressive_rendering": "VoilaConfiguration.progressive_rendering",
         "strip_sources": "VoilaConfiguration.strip_sources",
         "template": "VoilaConfiguration.template",
         "theme": "VoilaConfiguration.theme",
         "classic_tree": "VoilaConfiguration.classic_tree",
         "kernel_spec_manager_class": "VoilaConfiguration.kernel_spec_manager_class",
     }
+    if JUPYTER_SERVER_2:
+        aliases = {**aliases, "token": "Voila.token"}
     classes = [VoilaConfiguration, VoilaExecutor, VoilaExporter]
     connection_dir_root = Unicode(
         config=True,
@@ -335,50 +357,61 @@ class Voila(Application):
         ),
     )
 
-    cookie_secret = Bytes(
-        b"",
-        config=True,
-        help="""The random bytes used to secure cookies.
-        By default this is a new random number every time you start the server.
-        Set it to a value in a config file to enable logins to persist across server sessions.
+    @validate("prelaunch_hook")
+    def _valid_prelaunch_hook(self, proposal):
+        warn(
+            "Voila.prelaunch_hook is deprecated, please use VoilaConfiguration.prelaunch_hook instead",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        self.voila_configuration.prelaunch_hook = proposal["value"]
+        return proposal["value"]
 
-        Note: Cookie secrets should be kept private, do not share config files with
-        cookie_secret stored in plaintext (you can read the value from a file).
-        """,
-    )
+    if JUPYTER_SERVER_2:
+        cookie_secret = Bytes(
+            b"",
+            config=True,
+            help="""The random bytes used to secure cookies.
+            By default this is a new random number every time you start the server.
+            Set it to a value in a config file to enable logins to persist across server sessions.
 
-    token = Unicode(None, help="""Token for identity provider """, allow_none=True).tag(
-        config=True
-    )
+            Note: Cookie secrets should be kept private, do not share config files with
+            cookie_secret stored in plaintext (you can read the value from a file).
+            """,
+        )
 
-    auto_token = Bool(
-        False, help="""Generate token automatically """, allow_none=True
-    ).tag(config=True)
+        token = Unicode(
+            None, help="""Token for identity provider """, allow_none=True
+        ).tag(config=True)
 
-    @default("cookie_secret")
-    def _default_cookie_secret(self):
-        return os.urandom(32)
+        auto_token = Bool(
+            False, help="""Generate token automatically """, allow_none=True
+        ).tag(config=True)
 
-    authorizer_class = Type(
-        default_value=AllowAllAuthorizer,
-        klass=Authorizer,
-        config=True,
-        help=_("The authorizer class to use."),
-    )
+        @default("cookie_secret")
+        def _default_cookie_secret(self):
+            return os.urandom(32)
 
-    identity_provider_class = Type(
-        default_value=PasswordIdentityProvider,
-        klass=IdentityProvider,
-        config=True,
-        help=_("The identity provider class to use."),
-    )
+        authorizer_class = Type(
+            default_value=AllowAllAuthorizer,
+            klass=Authorizer,
+            config=True,
+            help=_("The authorizer class to use."),
+        )
 
-    kernel_websocket_connection_class = Type(
-        default_value=ZMQChannelsWebsocketConnection,
-        klass=BaseKernelWebsocketConnection,
-        config=True,
-        help=_("The kernel websocket connection class to use."),
-    )
+        identity_provider_class = Type(
+            default_value=PasswordIdentityProvider,
+            klass=IdentityProvider,
+            config=True,
+            help=_("The identity provider class to use."),
+        )
+
+        kernel_websocket_connection_class = Type(
+            default_value=ZMQChannelsWebsocketConnection,
+            klass=BaseKernelWebsocketConnection,
+            config=True,
+            help=_("The kernel websocket connection class to use."),
+        )
 
     @property
     def display_url(self):
@@ -390,7 +423,7 @@ class Voila(Application):
             ip = "%s" % socket.gethostname() if self.ip in ("", "0.0.0.0") else self.ip
             url = self._url(ip)
         # TODO: do we want to have the token?
-        if self.identity_provider.token:
+        if JUPYTER_SERVER_2 and self.identity_provider.token:
             # Don't log full token if it came from config
             token = (
                 self.identity_provider.token
@@ -441,7 +474,7 @@ class Voila(Application):
 
     @property
     def labextensions_path(self):
-        return jupyter_path("labextensions")
+        return get_voila_labextensions_path()
 
     @property
     def data_dir(self):
@@ -547,7 +580,8 @@ class Voila(Application):
             self.static_paths = collect_static_paths(
                 ["voila", "nbconvert"], template_name
             )
-            self.static_paths.append(DEFAULT_STATIC_FILES_PATH)
+            if JUPYTER_SERVER_2:
+                self.static_paths.append(DEFAULT_STATIC_FILES_PATH)
             conf_paths = [os.path.join(d, "conf.json") for d in self.template_paths]
             for p in conf_paths:
                 # see if config file exists
@@ -592,13 +626,25 @@ class Voila(Application):
         preheat_kernel: bool = self.voila_configuration.preheat_kernel
         pool_size: int = self.voila_configuration.default_pool_size
 
-        if preheat_kernel and self.prelaunch_hook:
+        if preheat_kernel and self.voila_configuration.prelaunch_hook:
             raise Exception("`preheat_kernel` and `prelaunch_hook` are incompatible")
 
+        progressive_rendering = self.voila_configuration.progressive_rendering
+        if preheat_kernel and progressive_rendering:
+            raise Exception(
+                "`preheat_kernel` and `progressive_rendering` are incompatible"
+            )
+
+        if not JUPYTER_SERVER_2 and progressive_rendering:
+            raise Exception(
+                "`progressive_rendering` can only be enabled with jupyter_server>=2"
+            )
+
         kernel_manager_class = voila_kernel_manager_factory(
             self.voila_configuration.multi_kernel_manager_class,
             preheat_kernel,
             pool_size,
+            page_config_hook=self.voila_configuration.page_config_hook,
         )
         self.kernel_manager = kernel_manager_class(
             parent=self,
@@ -622,35 +668,36 @@ class Voila(Application):
             extensions=["jinja2.ext.i18n"],
             **jenv_opt,
         )
-        server_env = jinja2.Environment(
-            loader=jinja2.FileSystemLoader(DEFAULT_TEMPLATE_PATH_LIST),
-            extensions=["jinja2.ext.i18n"],
-            **jenv_opt,
-        )
-
         nbui = gettext.translation(
             "nbui", localedir=os.path.join(ROOT, "i18n"), fallback=True
         )
         env.install_gettext_translations(nbui, newstyle=False)
-        server_env.install_gettext_translations(nbui, newstyle=False)
 
-        identity_provider_kwargs = {
-            "parent": self,
-            "log": self.log,
-            "login_handler_class": VoilaLoginHandler,
-        }
-        if self.token is None and not self.auto_token:
-            identity_provider_kwargs["token"] = ""
-        elif self.token is not None:
-            identity_provider_kwargs["token"] = self.token
+        if JUPYTER_SERVER_2:
+            server_env = jinja2.Environment(
+                loader=jinja2.FileSystemLoader(DEFAULT_TEMPLATE_PATH_LIST),
+                extensions=["jinja2.ext.i18n"],
+                **jenv_opt,
+            )
+            server_env.install_gettext_translations(nbui, newstyle=False)
 
-        self.identity_provider = self.identity_provider_class(
-            **identity_provider_kwargs
-        )
+            identity_provider_kwargs = {
+                "parent": self,
+                "log": self.log,
+                "login_handler_class": VoilaLoginHandler,
+            }
+            if self.token is None and not self.auto_token:
+                identity_provider_kwargs["token"] = ""
+            elif self.token is not None:
+                identity_provider_kwargs["token"] = self.token
 
-        self.authorizer = self.authorizer_class(
-            parent=self, log=self.log, identity_provider=self.identity_provider
-        )
+            self.identity_provider = self.identity_provider_class(
+                **identity_provider_kwargs
+            )
+
+            self.authorizer = self.authorizer_class(
+                parent=self, log=self.log, identity_provider=self.identity_provider
+            )
 
         settings = dict(
             base_url=self.base_url,
@@ -660,20 +707,25 @@ class Voila(Application):
             allow_remote_access=True,
             autoreload=self.autoreload,
             voila_jinja2_env=env,
-            jinja2_env=server_env,
+            jinja2_env=server_env if JUPYTER_SERVER_2 else env,
             server_root_dir="/",
             contents_manager=self.contents_manager,
             config_manager=self.config_manager,
-            cookie_secret=self.cookie_secret,
-            authorizer=self.authorizer,
-            identity_provider=self.identity_provider,
-            kernel_websocket_connection_class=self.kernel_websocket_connection_class,
-            login_url=url_path_join(self.base_url, "/login"),
             mathjax_config=self.mathjax_config,
             mathjax_url=self.mathjax_url,
         )
         settings[self.name] = self  # Why???
 
+        if JUPYTER_SERVER_2:
+            settings = {
+                **settings,
+                "cookie_secret": self.cookie_secret,
+                "authorizer": self.authorizer,
+                "identity_provider": self.identity_provider,
+                "kernel_websocket_connection_class": self.kernel_websocket_connection_class,
+                "login_url": url_path_join(self.base_url, "/login"),
+            }
+
         return settings
 
     def init_handlers(self) -> List:
@@ -692,7 +744,7 @@ class Voila(Application):
                     url_path_join(
                         self.server_url, r"/api/kernels/%s/channels" % _kernel_id_regex
                     ),
-                    KernelWebsocketHandler,
+                    VoilaKernelWebsocketHandler,
                 ),
                 (
                     url_path_join(self.server_url, r"/voila/templates/(.*)"),
@@ -709,7 +761,7 @@ class Voila(Application):
                     {
                         "themes_url": "/voila/api/themes",
                         "path": self.themes_dir,
-                        "labextensions_path": jupyter_path("labextensions"),
+                        "labextensions_path": get_voila_labextensions_path(),
                         "no_cache_paths": ["/"],
                     },
                 ),
@@ -719,7 +771,10 @@ class Voila(Application):
                 ),
             ]
         )
-        handlers.extend(self.identity_provider.get_handlers())
+
+        if JUPYTER_SERVER_2:
+            handlers.extend(self.identity_provider.get_handlers())
+
         if self.voila_configuration.preheat_kernel:
             handlers.append(
                 (
@@ -729,6 +784,15 @@ class Voila(Application):
                     RequestInfoSocketHandler,
                 )
             )
+        if self.voila_configuration.progressive_rendering:
+            handlers.append(
+                (
+                    url_path_join(
+                        self.server_url, r"/voila/execution/%s" % _kernel_id_regex
+                    ),
+                    ExecutionRequestHandler,
+                )
+            )
         # Serving JupyterLab extensions
         handlers.append(
             (
@@ -764,7 +828,6 @@ class Voila(Application):
                         "template_paths": self.template_paths,
                         "config": self.config,
                         "voila_configuration": self.voila_configuration,
-                        "prelaunch_hook": self.prelaunch_hook,
                     },
                 )
             )
@@ -772,11 +835,15 @@ class Voila(Application):
             self.log.debug("serving directory: %r", self.root_dir)
             handlers.extend(
                 [
-                    (self.server_url, TornadoVoilaTreeHandler, tree_handler_conf),
+                    (
+                        self.server_url,
+                        TornadoVoilaTreeHandler,
+                        {"voila_configuration": self.voila_configuration},
+                    ),
                     (
                         url_path_join(self.server_url, r"/voila/tree" + path_regex),
                         TornadoVoilaTreeHandler,
-                        tree_handler_conf,
+                        {"voila_configuration": self.voila_configuration},
                     ),
                     (
                         url_path_join(self.server_url, r"/voila/render/(.*)"),
@@ -785,7 +852,6 @@ class Voila(Application):
                             "template_paths": self.template_paths,
                             "config": self.config,
                             "voila_configuration": self.voila_configuration,
-                            "prelaunch_hook": self.prelaunch_hook,
                         },
                     ),
                     # On serving a directory, expose the content handler.
@@ -840,6 +906,8 @@ class Voila(Application):
             try:
                 self.app.listen(port, self.ip)
                 self.port = port
+                if self.voila_configuration.progressive_rendering:
+                    self.log.info("Progressive rendering is activated")
                 self.log.info("Voilà is running at:\n%s" % self.display_url)
             except OSError as e:
                 if e.errno == errno.EADDRINUSE:
